import unittest
# from unittest.mock import MagicMock
from unittest import mock

import os
import json
import monitor_exporter.monitorconnection as Monitor
import monitor_exporter.fileconfiguration as fileconfig
import monitor_exporter.perfdata as perfdata

import asyncio


def get_perf_mock_file(*args, **kwargs):
    file = open(os.path.join('test_data', 'perfdata.json'), 'r')
    json_file = file.read()
    file.close()

    return json.loads(json_file)


def get_custom_vars_mock_file(*args, **kwargs):
    file = open(os.path.join('test_data', 'perflabels.json'), 'r')
    json_file = file.read()
    file.close()

    return json.loads(json_file)


class TestMonitor(unittest.TestCase):

    def setUp(self):
        self.config = fileconfig.read_config(os.path.join('test_data', 'config.yml'))

    def test_get_perfdata_from_monitor(self):
        mcon = Monitor.MonitorConfig(self.config)

        # Create Mock
        mcon.get_host_data = AsyncMock(return_value=get_perf_mock_file())

        perf = perfdata.Perfdata(mcon, 'google.se')

        # Get the data from Monitor
        monitor_data = _run(perf.get_perfdata())
        self.assertEqual(len(monitor_data), 11)

        self.assertTrue(
            'monitor_host_state{hostname="google.se", environment="production"}' in monitor_data)
        self.assertTrue(
            'monitor_check_disk_local_mb_slash_bytes{hostname="google.se", service="disk_root", environment="production"}' in monitor_data)
        self.assertTrue(
            'monitor_check_ping_rta_seconds{hostname="google.se", service="pingit", environment="production"}' in monitor_data)


def _run(coro):
    return asyncio.get_event_loop().run_until_complete(coro)


def AsyncMock(*args, **kwargs):
    m = mock.MagicMock(*args, **kwargs)

    async def mock_coro(*args, **kwargs):
        return m(*args, **kwargs)

    mock_coro.mock = m
    return mock_coro



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2019  Opsdis AB

    This file is part of monitor-exporter.

    monitor-exporter is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    monitor-exporter is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with monitor-exporter.  If not, see <http://www.gnu.org/licenses/>.

"""
import re

import urllib3

import monitor_exporter.monitorconnection as Monitor
import monitor_exporter.log as log

# Disable InsecureRequestWarning
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

HOSTNAME = 'hostname'
SERVICE = 'service'
NAN = 'NaN'

check_command_regex = re.compile(r'^.+?[^!\n]+')
metric_verify_regexp = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]*$')


class Perfdata:

    def __init__(self, monitor: Monitor, query_hostname: str):
        # Get Monitor configuration and build URL
        self.monitor = monitor
        self.query_hostname = query_hostname
        self.prefix = monitor.get_prefix()
        self.allow_nan = monitor.is_allow_nan()
        self.configured_labels = monitor.get_configured_labels()
        self.perfname_to_label = monitor.get_perfname_to_label()
        self.perfdatadict = {}

    async def get_perfdata(self):

        # Use prometheus_labels method to fetch extra labels
        host_data = await self.monitor.get_host_data(self.query_hostname)

        custom_vars = {}
        host_state = None
        host_check_command = None
        host_perf_data = None

        if 'custom_variables' in host_data:
            custom_vars = host_data['custom_variables']
        if 'state' in host_data:
            host_state = host_data['state']
        if 'perf_data' in host_data:
            host_perf_data = host_data['perf_data']
        if 'check_command' in host_data:
            host_check_command = check_command_regex.search(host_data['check_command']).group()

        host_custom_vars_labels = self.prometheus_labels(custom_vars)

        # Host state
        labels = {HOSTNAME: self.query_hostname}
        labels.update(host_custom_vars_labels)

        # Additional labels for downtime,
        if 'downtime' in host_data:
            labels['downtime'] = str(host_data['downtime']).lower()
        if 'is_flapping' in host_data:
            labels['flapping'] = str(host_data['is_flapping'])
        if 'address' in host_data:
            labels['address'] = host_data['address']
        if 'acknowledged' in host_data:
            labels['acknowledged'] = str(host_data['acknowledged'])

        if host_state is not None:
            normalized_value, prometheus_key_with_labels = self.create_metric('host', labels,
                                                                              'state',
                                                                              {'value': int(host_state)})
            self.perfdatadict.update({prometheus_key_with_labels: str(normalized_value)})

        if host_check_command and host_perf_data:
            labels.update({SERVICE: 'isalive'})
            for perf_data_key, perf_data_value in host_perf_data.items():
                normalized_value, prometheus_key_with_labels = self.create_metric(host_check_command, labels,
                                                                                  perf_data_key, perf_data_value)

                if normalized_value != NAN or self.allow_nan:
                    prometheus_key = prometheus_key_with_labels.split('{')[0]
                    if metric_verify_regexp.match(prometheus_key):
                        self.perfdatadict.update({prometheus_key_with_labels: str(normalized_value)})
                    else:
                        log.warn("Illegal prometheus key - dropping",
                                 {'host': self.query_hostname, 'check_command': host_check_command, 'prometheus_key': prometheus_key})
                else:
                    log.warn("Missing value - dropping",
                             {'host': self.query_hostname, 'check_command': host_check_command})

        service_state_histo = {
            'bucket': {'0': 0, '1': 0, '2': 0, '+Inf': 0},
            '_count': 0,
            '_sum': 0
        }

        service_data = []
        if 'services' in host_data:
            service_data = host_data['services']

        for item in service_data:
            check_command = check_command_regex.search(item['check_command']).group()
            labels = {HOSTNAME: item['host']['name'], SERVICE: item['description']}
            # Add the host custom variables
            labels.update(host_custom_vars_labels)

            if 'downtime' in host_data and bool(host_data['downtime']):
                labels['downtime'] = str(host_data['downtime']).lower()
            elif 'downtime' in item:
                labels['downtime'] = str(item['downtime']).lower()
            if 'address' in host_data:
                labels['address'] = host_data['address']
            if 'is_flapping' in item:
                labels['flapping'] = str(item['is_flapping'])
            if 'acknowledged' in item:
                labels['acknowledged'] = str(item['acknowledged'])

            # For state if exists 0 OK, 1 Warning and 2 Critical
            if 'state' in item:
                normalized_value, prometheus_key_with_labels = self.create_metric_state(labels,
                                                                                        {'value': int(item['state'])})
                self.perfdatadict.update({prometheus_key_with_labels: str(normalized_value)})

                if int(item['state']) == 0:
                    service_state_histo['bucket']['0'] += 1
                elif int(item['state']) == 1:
                    service_state_histo['bucket']['1'] += 1
                elif int(item['state']) == 2:
                    service_state_histo['bucket']['2'] += 1
                else:
                    service_state_histo['bucket']['+Inf'] += 1
                service_state_histo['_count'] += 1
                service_state_histo['_sum'] += int(item['state'])

            if 'perf_data' in item and item['perf_data']:

                perfdata = item['perf_data']

                # For each perfname in perfdata
                for perf_data_key, perf_data_value in perfdata.items():
                    # get the value and unit

                    normalized_value, prometheus_key_with_labels = self.create_metric(check_command, labels,
                                                                                      perf_data_key,
                                                                                      perf_data_value)

                    if normalized_value != NAN or self.allow_nan:
                        prometheus_key = prometheus_key_with_labels.split('{')[0]
                        if metric_verify_regexp.match(prometheus_key):
                            self.perfdatadict.update({prometheus_key_with_labels: str(normalized_value)})
                        else:
                            log.warn("Illegal prometheus key - dropping",
                                     {'host': self.query_hostname, 'service': item['description'],
                                      'check_command': check_command, 'prometheus_key': prometheus_key})
                    else:
                        log.warn("Missing value - dropping",
                                 {'host': self.query_hostname, 'service': item['description'],
                                  'check_command': check_command})
        # self.create_service_state_histogram(service_state_histo, labels={HOSTNAME: self.query_hostname})

        return self.perfdatadict

    def create_service_state_histogram(self, histo: dict, labels: dict):
        name = 'service_state_by_host_bucket'
        for bucket, count in histo['bucket'].items():
            bucket_label = labels
            bucket_label['le'] = f"{bucket}"
            mertic_name = self.prefix + name + '{' + Perfdata.labels_string(labels) + '}'
            self.perfdatadict.update({mertic_name: str(count)})

        labels.pop('le')
        mertic_name = self.prefix + name + '_count{' + Perfdata.labels_string(labels) + '}'
        self.perfdatadict.update({mertic_name: str(histo['_count'])})
        mertic_name = self.prefix + name + '_sum{' + Perfdata.labels_string(labels) + '}'
        self.perfdatadict.update({mertic_name: str(histo['_sum'])})

    def create_metric(self, check_command, labels, perf_data_key, perf_data_value):

        perf_unit, perf_value, perf_warn, perf_crit = Perfdata.get_perfdata_value_unit(perf_data_value)

        normilized_value, unit = Perfdata.normalize_to_unit(perf_value, perf_unit)
        prometheus_key = self.get_metrics_name(check_command, perf_data_key, unit)
        if check_command in self.perfname_to_label:
            labels.update(
                Perfdata.add_labels_by_items(self.perfname_to_label[check_command]['label_name'],
                                             perf_data_key))
        prometheus_key_with_labels = Perfdata.concat_metrics_name_and_labels(labels, prometheus_key)
        return normilized_value, prometheus_key_with_labels

    def create_metric_state(self, labels, state_value):

        perf_unit, perf_value, perf_warn, perf_crit = Perfdata.get_perfdata_value_unit(state_value)

        normalized_value, unit = Perfdata.normalize_to_unit(perf_value, perf_unit)
        prometheus_key = self.prefix + 'service_state'

        prometheus_key_with_labels = Perfdata.concat_metrics_name_and_labels(labels, prometheus_key)
        return normalized_value, prometheus_key_with_labels

    @staticmethod
    def get_perfdata_value_unit(value: dict) -> tuple:
        perf_value = ''
        perf_unit = ''
        perf_warn = ''
        perf_crit = ''
        if 'value' in value:
            perf_value = value['value']
        if 'unit' in value:
            perf_unit = value['unit']
        if 'warn' in value:
            perf_warn = value['warn']
        if 'crit' in value:
            perf_crit = value['crit']

        return perf_unit, perf_value, perf_warn, perf_crit

    def get_metrics_name(self, check_command, key, unit):
        if unit:
            if check_command in self.perfname_to_label:
                prometheus_key = self.prefix + check_command + '_' + unit
            else:
                prometheus_key = self.prefix + check_command + '_' + key.lower() + '_' + unit
        else:
            if check_command in self.perfname_to_label:
                prometheus_key = self.prefix + check_command
            else:
                prometheus_key = self.prefix + check_command + '_' + key.lower()
        prometheus_key = Perfdata.rem_illegal_chars(prometheus_key)
        return prometheus_key

    def prometheus_labels(self, monitor_custom_vars):
        """
        Extract metric labels from custom_vars.
        Only defined custom vars are selected - see config.yml

        :return:
        custom_vars = {}
        for var in custom_vars_json:
            custom_vars = var['custom_variables']

        """

        new_labels = {}

        if monitor_custom_vars:
            # Make all variables to lower
            monitor_custom_vars = {k.lower(): v for k, v in monitor_custom_vars.items()}

            # Translate if configured
            for key, value in monitor_custom_vars.items():
                if key in self.configured_labels:
                    new_labels.update({self.configured_labels[key]: value})
                elif self.monitor.is_all_custom_vars():
                    new_labels.update({key: value})

        return new_labels

    def prometheus_format(self):
        # Build prometheus formatted data
        metrics = ''
        for key, value in self.perfdatadict.items():
            metrics += key + ' ' + value + '\n'
        return metrics

    @staticmethod
    def rem_illegal_chars(prometheus_key):
        # Replace illegal characters in metric name
        prometheus_key = prometheus_key.replace(' ', '_')
        prometheus_key = prometheus_key.replace('-', '_')
        prometheus_key = prometheus_key.replace('/', 'slash')
        prometheus_key = prometheus_key.replace('%', 'percent')
        prometheus_key = prometheus_key.replace('#', 'hash')
        prometheus_key = prometheus_key.replace('@', 'at')
        prometheus_key = prometheus_key.replace('$', 'dollar')
        prometheus_key = prometheus_key.replace('(', '_')
        prometheus_key = prometheus_key.replace(')', '_')
        prometheus_key = prometheus_key.replace('.', '_')
        prometheus_key = prometheus_key.replace('[', '_')
        prometheus_key = prometheus_key.replace(']', '_')
        prometheus_key = prometheus_key.replace('\\', '_')
        prometheus_key = prometheus_key.replace(':', '_')
        return prometheus_key

    @staticmethod
    def normalize_to_unit(value, unit):
        """Normalize the value to the unit returned.
        We use base-1000 for second-based units, and base-1024 for
        byte-based units. Sadly, the Nagios-Plugins specification doesn't
        disambiguate base-1000 (KB) and base-1024 (KiB).
        """
        if isinstance(value, str):
            return NAN, ''
        if unit == '%':
            return value / 100, 'ratio'
        if unit == 's':
            return value, 'seconds'
        if unit == 'ms':
            return value / 1000.0, 'seconds'
        if unit == 'us':
            return value / 1000000.0, 'seconds'
        if unit == 'B':
            return value, 'bytes'
        if unit == 'KB':
            return value * 1024, 'bytes'
        if unit == 'MB':
            return value * 1024 * 1024, 'bytes'
        if unit == 'GB':
            return value * 1024 * 1024 * 1024, 'bytes'
        if unit == 'TB':
            return value * 1024 * 1024 * 1024 * 1024, 'bytes'

        return value, ''

    @staticmethod
    def concat_metrics_name_and_labels(labels: dict, prometheus_key: str) -> str:
        """
        Build metric name with labels like
        metrics_name{label1="value1, .... }
        :param labels:
        :param prometheus_key:
        :return:
        """

        labelstring = Perfdata.labels_string(labels)
        prometheus_key = prometheus_key + '{' + labelstring + '}'

        return prometheus_key

    @staticmethod
    def labels_string(labels: dict) -> str:
        """
        Create a comma separated string of
        labels1=value1, ....
        :param labels:
        :return:
        """
        labelstring = ''
        sep = ''
        for label_key, label_value in labels.items():
            # Can only add custom vars that are simple strings. In incinga these can be complex dict structures
            # if type(label_value) is str or type(label_value) is int:
            if type(label_value) is str:
                labelstring += sep + label_key + '="' + label_value + '"'
                sep = ', '
        return labelstring

    @staticmethod
    def add_labels_by_items(label: str, key: str) -> dict:
        item_label = {label.lower(): key}
        return item_label



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2019  Opsdis AB

    This file is part of monitor-exporter.

    monitor-exporter is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    monitor-exporter is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with monitor-exporter.  If not, see <http://www.gnu.org/licenses/>.

"""

import argparse
import asyncio
from quart import Quart
from apscheduler.schedulers.asyncio import AsyncIOScheduler


import monitor_exporter.fileconfiguration as config
import monitor_exporter.log as log
import monitor_exporter.monitorconnection as monitorconnection
import monitor_exporter.proxy as proxy

default_interval = 60
default_ttl = 300


def start():
    """
    Used from __main__ to start as simple flask app
    :return:
    """
    parser = argparse.ArgumentParser(description='monitor_exporter')

    parser.add_argument('-f', '--configfile',
                        dest="configfile", help="configuration file")

    parser.add_argument('-p', '--port',
                        dest="port", help="Server port")

    args = parser.parse_args()

    port = 9631

    config_file = 'config.yml'
    if args.configfile:
        config_file = args.configfile

    configuration = config.read_config(config_file)
    if 'port' in configuration:
        port = configuration['port']

    if args.port:
        port = args.port

    log.configure_logger(configuration)

    monitorconnection.MonitorConfig(configuration)
    # Need to create an event loop for apscheduler
    loop = asyncio.new_event_loop()
    # Set the event loop
    asyncio.set_event_loop(loop)

    start_scheduler(configuration)

    log.info(f"Starting web app on port {port}")

    app = Quart(__name__)

    app.register_blueprint(proxy.app, url_prefix='')

    # Use the existing event loop
    app.run(host='0.0.0.0', port=port, loop=loop)


def create_app(config_path=None):
    """
    Used typical from gunicorn if need to pass config file different from default, e.g.
    gunicorn -b localhost:5000 --access-logfile /dev/null -w 4 "wsgi:create_app('/tmp/config.yml')"
    :param config_path:
    :return:
    """
    config_file = 'config.yml'
    if config_path:
        config_file = config_path

    configuration = config.read_config(config_file)

    log.configure_logger(configuration)

    monitorconnection.MonitorConfig(configuration)

    start_scheduler(configuration)
    log.info('Starting web app')

    app = Quart(__name__)
    app.register_blueprint(proxy.app, url_prefix='')

    return app


def start_scheduler(configuration):
    if 'cache' in configuration:

        scheduler = AsyncIOScheduler()
        seconds = default_interval if configuration.get('cache').get('interval') is None \
            else configuration.get('cache').get('interval')
        ttl = default_ttl if configuration.get('cache').get('ttl') is None \
            else configuration.get('cache').get('ttl')
        log.info(f"Monitor collector will run every {seconds} sec")
        # Run once at start up
        monitorconnection.MonitorConfig().collect_cache(ttl)
        scheduler.add_job(monitorconnection.MonitorConfig().collect_cache, trigger='interval', args=[ttl],
                          seconds=seconds)
        scheduler.start()



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2019  Opsdis AB

    This file is part of monitor-exporter.

    monitor-exporter is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    monitor-exporter is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with monitor-exporter.  If not, see <http://www.gnu.org/licenses/>.

"""

import logging
import datetime
from pythonjsonlogger import jsonlogger
from quart.logging import serving_handler, default_handler

logger = logging.getLogger('monitor-exporter')


def configure_logger(config):
    log_filename, log_level = read_config(config)

    if log_filename:
        hdlr = logging.FileHandler(log_filename)
    else:
        hdlr = logging.StreamHandler()

    formatter = CustomJsonFormatter('%(timestamp)s %(level)s %(name)s %(message)s')

    # Set the quart log handlers
    serving_handler.setFormatter(formatter)
    serving_handler.setLevel(log_level)
    default_handler.setFormatter(formatter)
    default_handler.setLevel(log_level)

    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.setLevel(log_level)

    # Add our handler to all loggers
    root = logging.root
    existing = root.manager.loggerDict.keys()
    for log1 in [logging.getLogger(name) for name in existing]:
        log1.addHandler(hdlr)


def read_config(config):
    log_filename = None
    log_level = 'INFO'
    if 'logger' in config:
        if 'logfile' in config['logger']:
            log_filename = config['logger']['logfile']

        if 'level' in config['logger']:
            log_level = config['logger']['level']
    return log_filename, log_level


def error(message, json_dict=None):
    logit(logger.error, json_dict, message)


def warn(message, json_dict=None):
    logit(logger.warning, json_dict, message)


def info(message, json_dict=None):
    logit(logger.info, json_dict, message)


def debug(message, json_dict=None):
    logit(logger.debug, json_dict, message)


def logit(log_func, json_dict, message):
    if json_dict:
        log_func('{}'.format(message), extra=json_dict)
    else:
        log_func('{}'.format(message))


def info_response_time(message: str, r_time: float):
    response_time = {"response_time_seconds": r_time}
    logger.info('{}'.format(message), extra=response_time)


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        super(CustomJsonFormatter, self).add_fields(log_record, record, message_dict)
        if not log_record.get('timestamp'):
            # this doesn't use record.created, so it is slightly off
            now = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
            log_record['timestamp'] = now
        if log_record.get('level'):
            log_record['level'] = log_record['level'].upper()
        else:
            log_record['level'] = record.levelname



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2019  Opsdis AB

    This file is part of monitor-exporter.

    monitor-exporter is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    monitor-exporter is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with monitor-exporter.  If not, see <http://www.gnu.org/licenses/>.

"""
import asyncio

from prometheus_client import (CONTENT_TYPE_LATEST, Counter)
from quart import request, Response, jsonify, Blueprint

import monitor_exporter.log as log
import monitor_exporter.monitorconnection as monitorconnection
from monitor_exporter.perfdata import Perfdata

app = Blueprint("prom", __name__)
total_requests = Counter('requests', 'Total requests to monitor-exporter endpoint')


@app.route('/', methods=['GET'])
def hello_world():
    return 'monitor-exporter alive'


@app.route("/metrics", methods=['GET'])
async def get_metrics():
    before_request_func(request)
    target = request.args.get('target')

    log.info('Collect metrics', {'target': target})

    monitor_data = Perfdata(monitorconnection.MonitorConfig(), target)

    # Fetch performance data from Monitor
    await asyncio.get_event_loop().create_task(monitor_data.get_perfdata())

    target_metrics = monitor_data.prometheus_format()

    resp = Response(target_metrics)
    resp.headers['Content-Type'] = CONTENT_TYPE_LATEST

    return resp


@app.route("/health", methods=['GET'])
def get_health():
    return check_healthy()


def before_request_func(request):
    call_status = {'remote_addr': request.remote_addr, 'url': request.url}
    log.info('Access', call_status)


@app.after_request
def after_request_func(response):
    total_requests.inc()

    call_status = {'content_length': response.content_length, 'status': response.status_code,
                   'count': total_requests._value.get()}
    log.info('Response', call_status)

    return response


def check_healthy() -> Response:
    resp = jsonify({'status': 'ok'})
    resp.status_code = 200
    return resp



# -*- coding: utf-8 -*-
"""
    Copyright (C) 2019  Opsdis AB

    This file is part of monitor-exporter.

    monitor-exporter is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    monitor-exporter is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with monitor-exporter.  If not, see <http://www.gnu.org/licenses/>.

"""
import json
import time

import aiohttp
import redis
import requests
from requests.auth import HTTPBasicAuth
from typing import Dict

import monitor_exporter.log as log


class Singleton(type):
    """
    Provide singleton pattern to MonitorConfig. A new instance is only created if:
     - instance do not exists
     - config is provided in constructor call, __init__
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances or args:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class MonitorConfig(object, metaclass=Singleton):
    config_entry = 'op5monitor'
    # high number so all services is fetched
    default_limit = '10000'

    def __init__(self, config=None):
        """
        The constructor takes on single argument that is a config dict
        :param config:
        """
        self.is_cache = False
        if config:
            self.is_cache = False if 'cache' not in config else True

        if self.is_cache:
            self.redis_host = 'localhost' if config.get('cache').get('redis').get('host') is None else config.get(
                'cache').get('redis').get('host')
            self.redis_port = '6379' if config.get('cache').get('redis').get('port') is None else config.get(
                'cache').get('redis').get('port')
            self.redis_db = '0' if config.get('cache').get('redis').get('db') is None else config.get(
                'cache').get('redis').get('db')
            self.redis_auth = None if config.get('cache').get('redis').get('auth') is None else config.get(
                'cache').get('redis').get('auth')

        self.user = ''
        self.passwd = ''
        self.host = ''
        self.headers = {'content-type': 'application/json'}
        self.verify = False
        self.retries = 5
        self.timeout = 5
        self.prefix = ''
        self.labels = []
        self.url_query_service_data = ''
        self.perfname_to_label = []
        self.allow_all_custom_vars = False
        self.allow_nan = False

        if config:
            self.user = config[MonitorConfig.config_entry]['user']
            self.passwd = config[MonitorConfig.config_entry]['passwd']
            self.host = config[MonitorConfig.config_entry]['url']
            if 'metric_prefix' in config[MonitorConfig.config_entry]:
                self.prefix = config['op5monitor']['metric_prefix'] + '_'
            if 'host_custom_vars' in config[MonitorConfig.config_entry]:
                self.labels = config['op5monitor']['host_custom_vars']
            if 'perfnametolabel' in config[MonitorConfig.config_entry]:
                self.perfname_to_label = config[MonitorConfig.config_entry]['perfnametolabel']
            if 'timeout' in config[MonitorConfig.config_entry]:
                self.timeout = int(config[MonitorConfig.config_entry]['timeout'])
            if 'verify' in config[MonitorConfig.config_entry]:
                self.verify = bool(config[MonitorConfig.config_entry]['verify'])
            if 'all_custom_vars' in config[MonitorConfig.config_entry]:
                self.allow_all_custom_vars = bool(config[MonitorConfig.config_entry]['all_custom_vars'])
            if 'allow_nan' in config[MonitorConfig.config_entry]:
                self.allow_nan = bool(config[MonitorConfig.config_entry]['allow_nan'])

            # Collect service data for a single host
            self.url_query_service_data = self.host + \
                                          '/api/filter/query?query=[services]%20host.name="{}' \
                                          '"&columns=host.name,description,perf_data,check_command,state,' \
                                          'downtimes,acknowledged,is_flapping' \
                                          '&limit=' + self.default_limit

            # Collect host data for a single host
            self.url_get_host = self.host + \
                                '/api/filter/query?query=[hosts]%20name="{}' \
                                '"&columns=address,custom_variables,perf_data,check_command,state,' \
                                'downtimes,acknowledged,is_flapping'

            # Collect service data for all services - used in cache mode
            self.url_query_all_service_data = self.host + \
                                              '/api/filter/{}?query=[services]%20all' \
                                              '&columns=host.name,description,perf_data,check_command,state,' \
                                              'downtimes,acknowledged,is_flapping'

            # # Collect host data for all hosts - used in cache mode
            self.url_query_all_host = self.host + \
                                      '/api/filter/{}?query=[hosts]%20all' \
                                      '&columns=name,address,custom_variables,perf_data,check_command,state,' \
                                      'downtimes,acknowledged,is_flapping'

            self.url_downtime = self.host + \
                                '/api/filter/{}?query=[downtimes]%20all' \
                                '&columns=id,start_time,end_time,fixed'

    def get_user(self) -> str:
        return self.user

    def get_passwd(self) -> str:
        return self.passwd

    def get_header(self) -> Dict[str, str]:
        return self.headers

    def get_verify(self) -> bool:
        return self.verify

    def get_timeout(self) -> int:
        return self.timeout

    def get_url(self) -> str:
        return self.host

    def number_of_retries(self) -> int:
        return self.retries

    def get_prefix(self) -> str:
        return self.prefix

    def is_all_custom_vars(self) -> bool:
        return self.allow_all_custom_vars

    def is_allow_nan(self) -> bool:
        return self.allow_nan

    def get_configured_labels(self):
        labeldict = {}

        for label in self.labels:
            for custom_var, value in label.items():
                for key, prom_label in value.items():
                    labeldict.update({custom_var: prom_label})
        return labeldict

    def get_perfname_to_label(self):
        return self.perfname_to_label

    async def get_host_data(self, hostname):
        """
        Build new URL and get custom_vars from Monitor
        :param hostname:
        :return:
        """

        if self.is_cache:
            host_data = await self.get_cache_host_data(hostname)
        else:
            host_data_list = await self.get(self.url_get_host.format(hostname))
            service_data = await self.get(self.url_query_service_data.format(hostname))
            if host_data_list:
                host_data = host_data_list[0]
                host_data['services'] = service_data
            else:
                host_data = list()

        return host_data

    def get_sync(self, url):
        data_json = {}

        try:
            data_from_monitor = requests.get(url, auth=HTTPBasicAuth(self.user, self.passwd),
                                             verify=self.get_verify(), headers=self.get_header(),
                                             timeout=self.get_timeout())
            data_from_monitor.raise_for_status()

            log.debug('API call: ' + data_from_monitor.url)
            if data_from_monitor.status_code != 200:
                log.info("Response", {'status': data_from_monitor.status_code, 'error': data_json['error'],
                                      'full_error': data_json['full_error']})
            else:
                data_json = json.loads(data_from_monitor.content)
                log.info("call api {}".format(url), {'status': data_from_monitor.status_code,
                                                     'response_time': data_from_monitor.elapsed.total_seconds()})
        except requests.exceptions.RequestException as err:
            log.error("{}".format(str(err)))
            raise err

        return data_json

    async def get(self, url):

        async with aiohttp.ClientSession() as session:
            async with session.get(url, auth=aiohttp.BasicAuth(self.user, self.passwd),
                                   verify_ssl=False,
                                   headers=self.get_header()) as response:
                re = await response.text()
                return json.loads(re)

    async def get_cache_service_data(self, hostname):
        r = self.get_cache_connection()

        data = r.get(self.key_services(hostname))
        if data:
            return json.loads(data)
        else:
            return []

    async def get_cache_host_data(self, host_name):
        r = self.get_cache_connection()

        data = r.get(self.key_hosts(host_name))
        if data:
            return json.loads(data)
        else:
            return {}

    def collect_cache(self, ttl: int = 300):
        """
        Collect Monitor data for all objects and store in cache
        :param ttl:
        :return:
        """
        try:
            # Get downtime
            now = int(time.time())
            ongoing_downtime = set()
            # Get the count to know how many to query
            count_downtimes = self.get_sync(self.url_downtime.format('count'))
            if 'count' in count_downtimes and int(count_downtimes['count']) > 0:
                count = count_downtimes['count']

                downtimes = self.get_sync(self.url_downtime.format('query') + '&limit=' + str(count))

                for downtime in downtimes:

                    if downtime['start_time'] <= now <= downtime['end_time']:
                        # downtime['id'] is an int -> make it to a str to compare in set
                        ongoing_downtime.add(downtime['id'])

            # Get the service data from Monitor
            start_time = time.time()
            count_services = self.get_sync(self.url_query_all_service_data.format('count'))

            hosts_to_services = {}

            services_flat = []
            if 'count' in count_services:
                count = count_services['count']
                services_flat = self.get_sync(self.url_query_all_service_data.format('query') + '&limit=' + str(count))
                for service_item in services_flat:
                    if service_item['host']['name'] not in hosts_to_services:
                        hosts_to_services[service_item['host']['name']] = []
                    host_name = service_item['host']['name']
                    downtime = set(service_item.pop('downtimes'))
                    if downtime & ongoing_downtime:
                        service_item['downtime'] = True
                    else:
                        service_item['downtime'] = False

                    hosts_to_services[host_name].append(service_item)

            # Get the host data from Monitor
            count_hosts = self.get_sync(self.url_query_all_host.format('count'))
            hosts = []
            if 'count' in count_hosts:
                count = count_hosts['count']
                hosts = self.get_sync(self.url_query_all_host.format('query') + '&limit=' + str(count))

            start_redis_time = time.time()
            # Save host objects
            hosts_set = set()
            r = self.get_cache_connection()
            p = r.pipeline()
            for host in hosts:
                host_name = host['name']
                downtime = set(host.pop('downtimes'))
                if downtime & ongoing_downtime:
                    host['downtime'] = True
                else:
                    host['downtime'] = False

                hosts_set.add(host_name)
                if host_name in hosts_to_services:
                    host['services'] = hosts_to_services[host_name]
                p.set(self.key_hosts(host_name), json.dumps(host))
                p.expire(self.key_hosts(host_name), ttl)
            p.execute()

            # Build host index
            r = self.get_cache_connection()
            existing_hosts = r.smembers(self.key_host_index())
            log.info(f"Existing hosts {len(existing_hosts)}")
            log.info(f"Monitor hosts {len(hosts_set)}")

            del_hosts = existing_hosts - hosts_set
            add_hosts = hosts_set - existing_hosts
            log.info(f"Delete hosts {len(del_hosts)}")
            log.info(f"Add hosts {len(add_hosts)}")

            existing_downtimes = set(map(int, r.smembers(self.key_downtime_index())))
            log.info(f"Existing downtimes {len(existing_downtimes)}")
            log.info(f"Monitor downtimes {len(ongoing_downtime)}")
            del_downtimes = existing_downtimes - ongoing_downtime
            add_downtimes = ongoing_downtime - existing_downtimes
            log.info(f"Delete downtimes {len(del_downtimes)}")
            log.info(f"Add downtimes {len(add_downtimes)}")

            p = r.pipeline()
            for host in add_hosts:
                p.sadd(self.key_host_index(), host)
            for host in del_hosts:
                p.srem(self.key_host_index(), host)

            for downtime in add_downtimes:
                p.sadd(self.key_downtime_index(), downtime)
            for downtime in del_downtimes:
                p.srem(self.key_downtime_index(), downtime)

            p.expire(self.key_host_index(), ttl)
            p.expire(self.key_downtime_index(), ttl)
            p.execute()

            end_time = time.time()

            log.info(
                f"Monitor collector exec time total {(end_time - start_time)} "
                f"redis write {len(services_flat) + len(hosts)} objects in {end_time - start_redis_time}")
        except Exception as err:
            log.error(
                f"Monitor collector failed with {str(err)}")

    def key_services(self, host):
        return host + ':services'

    def key_hosts(self, host_name):
        return host_name + ":host"

    def key_host_index(self) -> str:
        return "hosts"

    def key_downtime_index(self) -> str:
        return "downtimes"

    def get_cache_connection(self):
        return redis.Redis(host=self.redis_host, port=self.redis_port, db=self.redis_db, password=self.redis_auth,
                           decode_responses=True)



from os.path import dirname, join
from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

def read(fname):
    return open(join(dirname(__file__), fname)).read()


setup(
    name='monitor-exporter',
    version_config={
        "template": "{tag}",
        "dev_template": "{tag}.dev{ccount}",
        "dirty_template": "{tag}.post{ccount}+git.{sha}.dirty",
        "starting_version": "0.0.1",
        "version_callback": None,
        "version_file": None,
        "count_commits_from_version_file": False,
        "branch_formatter": None
    },
    setup_requires=['setuptools-git-versioning'],
    packages=find_packages(),
    author='thenodon',
    author_email='anders@opsdis.com',
    url='https://github.com/opsdis/monitor-exporter',
    license='GPLv3',
    include_package_data=True,
    zip_safe=False,
    description='A Prometheus exporter for OP5 Monitor',
    install_requires=read('requirements.txt').split(),
    python_requires='>=3.6',
    long_description=long_description,
    long_description_content_type='text/markdown'
)



import asyncio
import os
import unittest
from unittest import mock

import monitor_exporter.fileconfiguration as fileconfig
import monitor_exporter.monitorconnection as Monitor


def mocked_monitorconnection_get(*args, **kwargs):
    yield {"TEST": "test"}


class TestMonitor(unittest.TestCase):

    def setUp(self):
        self.config = fileconfig.read_config(os.path.join('test_data', 'config.yml'))

    def test_default_config(self):
        mcon = Monitor.MonitorConfig()
        self.assertEqual(mcon.get_url(), '')

    def test_with_config(self):
        mcon = Monitor.MonitorConfig(self.config)
        self.assertEqual(mcon.get_url(), 'https://monitor.xyz')

    def test_singleton(self):
        mcon1 = Monitor.MonitorConfig(self.config)
        mcon2 = Monitor.MonitorConfig()
        self.assertEqual(mcon2.get_url(), 'https://monitor.xyz')
        # Make sure it same object - singleton
        self.assertEqual(id(mcon1), id(mcon2))
        mcon2 = Monitor.MonitorConfig(self.config)
        self.assertNotEqual(id(mcon1), id(mcon2))

    @mock.patch('monitor_exporter.monitorconnection.MonitorConfig.get', side_effect=mocked_monitorconnection_get())
    def test_get_perfdata_from_monitor(self, mock_get):
        mcon = Monitor.MonitorConfig(self.config)
        response = _run(mcon.get('sunet.se'))
        self.assertEqual(response, {"TEST": "test"})


def _run(coro):
    return asyncio.get_event_loop().run_until_complete(coro)
